@@ -6,7 +6,7 @@
  *
  ******************************************************************************
  * Copyright (c) 2013, Paul Ramsey <pramsey@boundlessgeo.com>
- * Copyright (c) 2014, Even Rouault <even dot rouault at mines-paris dot org>
+ * Copyright (c) 2014, Even Rouault <even dot rouault at spatialys.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -32,7 +32,7 @@
 #include "ogrsqliteutility.h"
 #include "ogr_p.h"
 
-CPL_CVSID("$Id: ogrgeopackagelayer.cpp 39983 2017-09-02 20:24:15Z rouault $");
+CPL_CVSID("$Id: ogrgeopackagelayer.cpp 61d987943110d648f8c231783559e48a08303ff3 2020-10-17 20:56:24 +0200 Even Rouault $")
 
 /************************************************************************/
 /*                      OGRGeoPackageLayer()                            */
@@ -40,14 +40,14 @@ CPL_CVSID("$Id: ogrgeopackagelayer.cpp 39983 2017-09-02 20:24:15Z rouault $");
 
 OGRGeoPackageLayer::OGRGeoPackageLayer(GDALGeoPackageDataset *poDS) :
     m_poDS(poDS),
-    m_poFeatureDefn(NULL),
+    m_poFeatureDefn(nullptr),
     iNextShapeId(0),
-    m_poQueryStatement(NULL),
+    m_poQueryStatement(nullptr),
     bDoStep(true),
-    m_pszFidColumn(NULL),
+    m_pszFidColumn(nullptr),
     iFIDCol(-1),
     iGeomCol(-1),
-    panFieldOrdinals(NULL)
+    panFieldOrdinals(nullptr)
 {}
 
 /************************************************************************/
@@ -77,6 +77,7 @@ void OGRGeoPackageLayer::ResetReading()
 {
     ClearStatement();
     iNextShapeId = 0;
+    m_bEOF = false;
 }
 
 /************************************************************************/
@@ -86,11 +87,11 @@ void OGRGeoPackageLayer::ResetReading()
 void OGRGeoPackageLayer::ClearStatement()
 
 {
-    if( m_poQueryStatement != NULL )
+    if( m_poQueryStatement != nullptr )
     {
         CPLDebug( "GPKG", "finalize %p", m_poQueryStatement );
         sqlite3_finalize( m_poQueryStatement );
-        m_poQueryStatement = NULL;
+        m_poQueryStatement = nullptr;
     }
 }
 
@@ -101,15 +102,18 @@ void OGRGeoPackageLayer::ClearStatement()
 OGRFeature *OGRGeoPackageLayer::GetNextFeature()
 
 {
+    if( m_bEOF )
+        return nullptr;
+
+    if( m_poQueryStatement == nullptr )
+    {
+        ResetStatement();
+        if (m_poQueryStatement == nullptr)
+            return nullptr;
+    }
+
     for( ; true; )
     {
-        if( m_poQueryStatement == NULL )
-        {
-            ResetStatement();
-            if (m_poQueryStatement == NULL)
-                return NULL;
-        }
-
     /* -------------------------------------------------------------------- */
     /*      Fetch a record (unless otherwise instructed)                    */
     /* -------------------------------------------------------------------- */
@@ -127,8 +131,9 @@ OGRFeature *OGRGeoPackageLayer::GetNextFeature()
                 }
 
                 ClearStatement();
+                m_bEOF = true;
 
-                return NULL;
+                return nullptr;
             }
         }
         else
@@ -138,9 +143,9 @@ OGRFeature *OGRGeoPackageLayer::GetNextFeature()
 
         OGRFeature *poFeature = TranslateFeature(m_poQueryStatement);
 
-        if( (m_poFilterGeom == NULL
+        if( (m_poFilterGeom == nullptr
             || FilterGeometry( poFeature->GetGeomFieldRef(m_iGeomFieldFilter) ) )
-            && (m_poAttrQuery == NULL
+            && (m_poAttrQuery == nullptr
                 || m_poAttrQuery->Evaluate( poFeature )) )
             return poFeature;
 
@@ -166,7 +171,7 @@ OGRFeature *OGRGeoPackageLayer::TranslateFeature( sqlite3_stmt* hStmt )
     if( iFIDCol >= 0 )
     {
         poFeature->SetFID( sqlite3_column_int64( hStmt, iFIDCol ) );
-        if( m_pszFidColumn == NULL && poFeature->GetFID() == 0 )
+        if( m_pszFidColumn == nullptr && poFeature->GetFID() == 0 )
         {
             // Miht be the case for views with joins.
             poFeature->SetFID( iNextShapeId );
@@ -192,8 +197,8 @@ OGRFeature *OGRGeoPackageLayer::TranslateFeature( sqlite3_stmt* hStmt )
             int iGpkgSize = sqlite3_column_bytes(hStmt, iGeomCol);
             // coverity[tainted_data_return]
             GByte *pabyGpkg = (GByte *)sqlite3_column_blob(hStmt, iGeomCol);
-            OGRGeometry *poGeom = GPkgGeometryToOGR(pabyGpkg, iGpkgSize, NULL);
-            if ( poGeom == NULL )
+            OGRGeometry *poGeom = GPkgGeometryToOGR(pabyGpkg, iGpkgSize, nullptr);
+            if ( poGeom == nullptr )
             {
                 // Try also spatialite geometry blobs
                 if( OGRSQLiteLayer::ImportSpatiaLiteGeometry( pabyGpkg, iGpkgSize,
@@ -202,7 +207,7 @@ OGRFeature *OGRGeoPackageLayer::TranslateFeature( sqlite3_stmt* hStmt )
                     CPLError( CE_Failure, CPLE_AppDefined, "Unable to read geometry");
                 }
             }
-            if( poGeom != NULL )
+            if( poGeom != nullptr )
                 poGeom->assignSpatialReference(poSrs);
             poFeature->SetGeometryDirectly( poGeom );
         }
@@ -219,7 +224,8 @@ OGRFeature *OGRGeoPackageLayer::TranslateFeature( sqlite3_stmt* hStmt )
 
         const int iRawField = panFieldOrdinals[iField];
 
-        if( sqlite3_column_type( hStmt, iRawField ) == SQLITE_NULL )
+        const int nSqlite3ColType = sqlite3_column_type( hStmt, iRawField );
+        if( nSqlite3ColType == SQLITE_NULL )
         {
             poFeature->SetFieldNull( iField );
             continue;
@@ -255,19 +261,111 @@ OGRFeature *OGRGeoPackageLayer::TranslateFeature( sqlite3_stmt* hStmt )
 
             case OFTDate:
             {
-                const char* pszTxt = (const char*)sqlite3_column_text( hStmt, iRawField );
-                int nYear, nMonth, nDay;
-                if( sscanf(pszTxt, "%d-%d-%d", &nYear, &nMonth, &nDay) == 3 )
-                    poFeature->SetField(iField, nYear, nMonth, nDay, 0, 0, 0, 0);
+                if( nSqlite3ColType == SQLITE_TEXT )
+                {
+                    const char* pszTxt = (const char*)sqlite3_column_text( hStmt, iRawField );
+                    int nYear, nMonth, nDay;
+                    if( sscanf(pszTxt, "%d-%d-%d", &nYear, &nMonth, &nDay) == 3 )
+                    {
+                        poFeature->SetField(iField, nYear, nMonth, nDay, 0, 0, 0, 0);
+                    }
+                    else if ( sscanf(pszTxt, "%d/%d/%d", &nYear, &nMonth, &nDay) == 3 )
+                    {
+                        poFeature->SetField(iField, nYear, nMonth, nDay, 0, 0, 0, 0);
+                        constexpr int line = __LINE__;
+                        if( !m_poDS->m_oSetGPKGLayerWarnings[line] )
+                        {
+                            CPLError(CE_Warning, CPLE_AppDefined,
+                                     "Non-conformant content for record "
+                                     CPL_FRMT_GIB " in column %s, %s, "
+                                     "successfully parsed",
+                                     poFeature->GetFID(),
+                                     poFieldDefn->GetNameRef(), pszTxt);
+                            m_poDS->m_oSetGPKGLayerWarnings[line] = true;
+                        }
+                    }
+                    else
+                    {
+                        constexpr int line = __LINE__;
+                        if( !m_poDS->m_oSetGPKGLayerWarnings[line] )
+                        {
+                            CPLError(CE_Warning, CPLE_AppDefined,
+                                     "Invalid content for record "
+                                     CPL_FRMT_GIB " in column %s: %s",
+                                     poFeature->GetFID(),
+                                     poFieldDefn->GetNameRef(), pszTxt);
+                            m_poDS->m_oSetGPKGLayerWarnings[line] = true;
+                        }
+                    }
+                }
+                else
+                {
+                    constexpr int line = __LINE__;
+                    if( !m_poDS->m_oSetGPKGLayerWarnings[line] )
+                    {
+                        CPLError(CE_Warning, CPLE_AppDefined,
+                                 "Unexpected data type for record "
+                                 CPL_FRMT_GIB " in column %s",
+                                 poFeature->GetFID(),
+                                 poFieldDefn->GetNameRef());
+                        m_poDS->m_oSetGPKGLayerWarnings[line] = true;
+                    }
+                }
                 break;
             }
 
             case OFTDateTime:
             {
-                const char* pszTxt = (const char*)sqlite3_column_text( hStmt, iRawField );
-                OGRField sField;
-                if( OGRParseXMLDateTime(pszTxt, &sField) )
-                    poFeature->SetField(iField, &sField);
+                if( nSqlite3ColType == SQLITE_TEXT )
+                {
+                    const char* pszTxt = (const char*)sqlite3_column_text( hStmt, iRawField );
+                    OGRField sField;
+                    if( OGRParseXMLDateTime(pszTxt, &sField) )
+                    {
+                        poFeature->SetField(iField, &sField);
+                    }
+                    else if ( OGRParseDate(pszTxt, &sField, 0) )
+                    {
+                        poFeature->SetField(iField, &sField);
+                        constexpr int line = __LINE__;
+                        if( !m_poDS->m_oSetGPKGLayerWarnings[line] )
+                        {
+                            CPLError(CE_Warning, CPLE_AppDefined,
+                                     "Non-conformant content for record "
+                                     CPL_FRMT_GIB " in column %s, %s, "
+                                     "successfully parsed",
+                                     poFeature->GetFID(),
+                                     poFieldDefn->GetNameRef(), pszTxt);
+                            m_poDS->m_oSetGPKGLayerWarnings[line] = true;
+                        }
+                    }
+                    else
+                    {
+                        constexpr int line = __LINE__;
+                        if( !m_poDS->m_oSetGPKGLayerWarnings[line] )
+                        {
+                            CPLError(CE_Warning, CPLE_AppDefined,
+                                     "Invalid content for record "
+                                     CPL_FRMT_GIB " in column %s: %s",
+                                     poFeature->GetFID(),
+                                     poFieldDefn->GetNameRef(), pszTxt);
+                           m_poDS->m_oSetGPKGLayerWarnings[line] = true;
+                        }
+                    }
+                }
+                else
+                {
+                    constexpr int line = __LINE__;
+                    if( !m_poDS->m_oSetGPKGLayerWarnings[line] )
+                    {
+                        CPLError(CE_Warning, CPLE_AppDefined,
+                                 "Unexpected data type for record "
+                                 CPL_FRMT_GIB " in column %s",
+                                 poFeature->GetFID(),
+                                 poFieldDefn->GetNameRef());
+                        m_poDS->m_oSetGPKGLayerWarnings[line] = true;
+                    }
+                }
                 break;
             }
 
@@ -305,7 +403,7 @@ int OGRGeoPackageLayer::TestCapability ( const char * pszCap )
     if( EQUAL(pszCap,OLCIgnoreFields) )
         return TRUE;
     else if ( EQUAL(pszCap, OLCStringsAsUTF8) )
-        return m_poDS->GetUTF8();
+        return TRUE;
     else
         return FALSE;
 }
@@ -342,10 +440,10 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
     {
         const char* pszTableName = sqlite3_column_table_name( hStmt, iCol );
         const char* pszOriginName = sqlite3_column_origin_name( hStmt, iCol );
-        if( pszTableName != NULL && pszOriginName != NULL )
+        if( pszTableName != nullptr && pszOriginName != nullptr )
         {
             OGRLayer* poLayer = m_poDS->GetLayerByName(pszTableName);
-            if( poLayer != NULL )
+            if( poLayer != nullptr )
             {
                 if( EQUAL(pszOriginName, poLayer->GetFIDColumn()) )
                 {
@@ -368,7 +466,7 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
         if( m_poFeatureDefn->GetFieldIndex( oField.GetNameRef() ) != -1 )
             continue;
 
-        if( m_pszFidColumn != NULL && EQUAL(m_pszFidColumn,
+        if( m_pszFidColumn != nullptr && EQUAL(m_pszFidColumn,
                                             oField.GetNameRef()) )
             continue;
 
@@ -389,10 +487,10 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
 #ifdef SQLITE_HAS_COLUMN_METADATA
         const char* pszTableName = sqlite3_column_table_name( hStmt, iCol );
         const char* pszOriginName = sqlite3_column_origin_name( hStmt, iCol );
-        if( pszTableName != NULL && pszOriginName != NULL )
+        if( pszTableName != nullptr && pszOriginName != nullptr )
         {
             OGRLayer* poLayer = m_poDS->GetLayerByName(pszTableName);
-            if( poLayer != NULL )
+            if( poLayer != nullptr )
             {
                 if( m_poFeatureDefn->GetGeomFieldCount() == 0 &&
                     EQUAL(pszOriginName, poLayer->GetGeometryColumn()) )
@@ -405,7 +503,7 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
                     continue;
                 }
                 else if( EQUAL(pszOriginName, poLayer->GetFIDColumn()) &&
-                         m_pszFidColumn == NULL && nFIDCandidates == 1 )
+                         m_pszFidColumn == nullptr && nFIDCandidates == 1 )
                 {
                     m_pszFidColumn = CPLStrdup(oField.GetNameRef());
                     iFIDCol = iCol;
@@ -431,7 +529,7 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
 #endif
 
         const int nColType = sqlite3_column_type( hStmt, iCol );
-        if( m_pszFidColumn == NULL && nColType == SQLITE_INTEGER &&
+        if( m_pszFidColumn == nullptr && nColType == SQLITE_INTEGER &&
             EQUAL(oField.GetNameRef(), "FID") )
         {
             m_pszFidColumn = CPLStrdup(oField.GetNameRef());
@@ -452,13 +550,13 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
                 const GByte* pabyGpkg =
                         (const GByte*)sqlite3_column_blob( hStmt, iCol  );
                 GPkgHeader oHeader;
-                OGRGeometry* poGeom = NULL;
+                OGRGeometry* poGeom = nullptr;
                 int nSRID = 0;
 
                 if( GPkgHeaderFromWKB(pabyGpkg, nBytes, &oHeader) ==
                                                                 OGRERR_NONE )
                 {
-                    poGeom = GPkgGeometryToOGR(pabyGpkg, nBytes, NULL);
+                    poGeom = GPkgGeometryToOGR(pabyGpkg, nBytes, nullptr);
                     nSRID = oHeader.iSrsId;
                 }
                 else
@@ -468,7 +566,7 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
                           pabyGpkg, nBytes, &poGeom, &nSRID ) != OGRERR_NONE )
                     {
                         delete poGeom;
-                        poGeom = NULL;
+                        poGeom = nullptr;
                     }
                 }
 
@@ -479,7 +577,7 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
 
                     /* Read the SRS */
                     OGRSpatialReference *poSRS =
-                                        m_poDS->GetSpatialRef(nSRID);
+                                        m_poDS->GetSpatialRef(nSRID, true);
                     if ( poSRS )
                     {
                         oGeomField.SetSpatialRef(poSRS);
@@ -487,7 +585,7 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
                     }
 
                     OGRwkbGeometryType eGeomType = poGeom->getGeometryType();
-                    if( pszDeclType != NULL )
+                    if( pszDeclType != nullptr )
                     {
                         OGRwkbGeometryType eDeclaredGeomType =
                             GPkgGeometryTypeToWKB(pszDeclType, false, false);
@@ -501,7 +599,7 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
                     oGeomField.SetType( eGeomType );
 
                     delete poGeom;
-                    poGeom = NULL;
+                    poGeom = nullptr;
 
                     m_poFeatureDefn->AddGeomFieldDefn(&oGeomField);
                     iGeomCol = iCol;
@@ -537,7 +635,7 @@ void OGRGeoPackageLayer::BuildFeatureDefn( const char *pszLayerName,
             /* leave it as OFTString */;
         }
 
-        if (pszDeclType != NULL)
+        if (pszDeclType != nullptr)
         {
             OGRFieldSubType eSubType;
             int nMaxWidth = 0;
